from conan.tools.env import Environment
from conans.model.build_info import DepCppInfo


class AutotoolsDeps:
    def __init__(self, conanfile):
        # Set the generic objects before mapping to env vars to let the user
        # alter some value
        self._conanfile = conanfile

        self.libs = []
        self.system_libs = []
        self.include_paths = []
        self.lib_paths = []
        self.defines = []
        self.cflags = []
        self.cxxflags = []
        self.sharedlinkflags = []
        self.exelinkflags = []
        self.frameworks = []
        self.framework_paths = []
        self.sysroot = None

        def merge_lists(seq1, seq2):
            return [s for s in seq1 if s not in seq2] + seq2

        def merge(dep):
            dep_cpp_info = DepCppInfo(dep.cpp_info)  # To deal with components
            self.system_libs = merge_lists(self.system_libs, dep_cpp_info.system_libs)
            self.include_paths = merge_lists(self.include_paths, dep_cpp_info.include_paths)
            self.lib_paths = merge_lists(self.lib_paths, dep_cpp_info.lib_paths)
            self.framework_paths = merge_lists(self.framework_paths, dep_cpp_info.framework_paths)
            self.libs = merge_lists(self.libs, dep_cpp_info.libs)
            self.frameworks = merge_lists(self.frameworks, dep_cpp_info.frameworks)

            # Note these are in reverse order
            self.defines = merge_lists(dep_cpp_info.defines, self.defines)
            self.cxxflags = merge_lists(dep_cpp_info.cxxflags, self.cxxflags)
            self.cflags = merge_lists(dep_cpp_info.cflags, self.cflags)
            self.sharedlinkflags = merge_lists(dep_cpp_info.sharedlinkflags, self.sharedlinkflags)
            self.exelinkflags = merge_lists(dep_cpp_info.exelinkflags, self.exelinkflags)

            if not self.sysroot:
                self.sysroot = dep_cpp_info.sysroot

        def _apply_transitive_runenv(next_requires):
            # TODO: This visitor is same as VirtualEnv runenv_info one, extract
            all_requires = []
            while next_requires:
                new_requires = []
                for require in next_requires:
                    # The explicit has more priority
                    merge(require)
                    all_requires.append(require)

                    for transitive in require.dependencies.requires:
                        # Avoid duplication/repetitions
                        if transitive not in new_requires and transitive not in all_requires:
                            new_requires.append(transitive)
                next_requires = new_requires

        _apply_transitive_runenv(self._conanfile.dependencies.requires)

    def environment(self):
        # cpp_flags
        cpp_flags = []
        include_paths = ['-I"%s"' % p for p in self.include_paths]
        cpp_flags.extend(include_paths)
        cpp_flags.extend(["-D%s" % define for define in self.defines])

        # Libs
        libs = ["-l%s" % library for library in self.libs]

        # Ldflags
        # TODO: Discuss, should the helper filter frameworks based on compiler?
        frameworks = ["-framework %s" % framework for framework in self.frameworks]
        frameworks_paths = ["-F %s" % framework_path for framework_path in self.framework_paths]
        ldflags = self.sharedlinkflags
        ldflags.extend(self.exelinkflags)
        ldflags.extend(frameworks)
        ldflags.extend(frameworks_paths)
        lib_paths = ['-L"%s"' % p for p in self.lib_paths]
        ldflags.extend(lib_paths)

        # cflags
        cflags = self.cflags
        cxxflags = self.cxxflags

        if self.sysroot:
            srf = '--sysroot={}'.format(self.sysroot)
            cflags.append(srf)
            cxxflags.append(srf)
            ldflags.append(srf)

        env = Environment()
        env.append("CPPFLAGS", cpp_flags)
        env.append("LIBS", libs)
        env.append("LDFLAGS", ldflags)
        env.append("CXXFLAGS", cxxflags)
        env.append("CFLAGS", cflags)
        return env

    def generate(self):
        env = self.environment()
        env.save_sh("conanautotoolsdeps.sh")
        env.save_bat("conanautotoolsdeps.bat")



from conans.client.tools.apple import to_apple_arch


def architecture_flag(settings):
    """
    returns flags specific to the target architecture and compiler
    """
    compiler = settings.get_safe("compiler")
    compiler_base = settings.get_safe("compiler.base")
    arch = settings.get_safe("arch")
    the_os = settings.get_safe("os")
    subsystem = settings.get_safe("os.subsystem")
    if not compiler or not arch:
        return ""

    if the_os == "Android":
        return ""

    if str(compiler) in ['gcc', 'apple-clang', 'clang', 'sun-cc']:
        if str(the_os) == 'Macos' and str(subsystem) == 'catalyst':
            # FIXME: This might be conflicting with Autotools --target cli arg
            apple_arch = to_apple_arch(arch)
            if apple_arch:
                return '--target=%s-apple-ios-macabi' % apple_arch
        elif str(arch) in ['x86_64', 'sparcv9', 's390x']:
            return '-m64'
        elif str(arch) in ['x86', 'sparc']:
            return '-m32'
        elif str(arch) in ['s390']:
            return '-m31'
        elif str(the_os) == 'AIX':
            if str(arch) in ['ppc32']:
                return '-maix32'
            elif str(arch) in ['ppc64']:
                return '-maix64'
    elif str(compiler) == "intel":
        # https://software.intel.com/en-us/cpp-compiler-developer-guide-and-reference-m32-m64-qm32-qm64
        if str(arch) == "x86":
            return "/Qm32" if str(compiler_base) == "Visual Studio" else "-m32"
        elif str(arch) == "x86_64":
            return "/Qm64" if str(compiler_base) == "Visual Studio" else "-m64"
    elif str(compiler) == "mcst-lcc":
        return {"e2k-v2": "-march=elbrus-v2",
                "e2k-v3": "-march=elbrus-v3",
                "e2k-v4": "-march=elbrus-v4",
                "e2k-v5": "-march=elbrus-v5",
                "e2k-v6": "-march=elbrus-v6",
                "e2k-v7": "-march=elbrus-v7"}.get(str(arch), "")
    return ""


def build_type_flags(settings):
    """
    returns flags specific to the build type (Debug, Release, etc.)
    (-s, -g, /Zi, etc.)
    """
    compiler = settings.get_safe("compiler.base") or settings.get_safe("compiler")

    build_type = settings.get_safe("build_type")
    vs_toolset = settings.get_safe("compiler.toolset")
    if not compiler or not build_type:
        return ""

    # https://github.com/Kitware/CMake/blob/d7af8a34b67026feaee558433db3a835d6007e06/
    # Modules/Platform/Windows-MSVC.cmake
    if str(compiler) == 'Visual Studio':
        if vs_toolset and "clang" in str(vs_toolset):
            flags = {"Debug": ["-gline-tables-only", "-fno-inline", "-O0"],
                     "Release": ["-O2"],
                     "RelWithDebInfo": ["-gline-tables-only", "-O2", "-fno-inline"],
                     "MinSizeRel": []
                     }.get(build_type, ["-O2", "-Ob2"])
        else:
            flags = {"Debug": ["-Zi", "-Ob0", "-Od"],
                     "Release": ["-O2", "-Ob2"],
                     "RelWithDebInfo": ["-Zi", "-O2", "-Ob1"],
                     "MinSizeRel": ["-O1", "-Ob1"],
                     }.get(build_type, [])
        return flags
    else:
        # https://github.com/Kitware/CMake/blob/f3bbb37b253a1f4a26809d6f132b3996aa2e16fc/
        # Modules/Compiler/GNU.cmake
        # clang include the gnu (overriding some things, but not build type) and apple clang
        # overrides clang but it doesn't touch clang either
        if str(compiler) in ["clang", "gcc", "apple-clang", "qcc", "mcst-lcc"]:
            # FIXME: It is not clear that the "-s" is something related with the build type
            # cmake is not adjusting it
            # -s: Remove all symbol table and relocation information from the executable.
            flags = {"Debug": ["-g"],
                     "Release": ["-O3", "-s"] if str(compiler) == "gcc" else ["-O3"],
                     "RelWithDebInfo": ["-O2", "-g"],
                     "MinSizeRel": ["-Os"],
                     }.get(build_type, [])
            return flags
        elif str(compiler) == "sun-cc":
            # https://github.com/Kitware/CMake/blob/f3bbb37b253a1f4a26809d6f132b3996aa2e16fc/
            # Modules/Compiler/SunPro-CXX.cmake
            flags = {"Debug": ["-g"],
                     "Release": ["-xO3"],
                     "RelWithDebInfo": ["-xO2", "-g"],
                     "MinSizeRel": ["-xO2", "-xspace"],
                     }.get(build_type, [])
            return flags
    return ""


def use_win_mingw(conanfile):
    if hasattr(conanfile, 'settings_build'):
        os_build = conanfile.settings_build.get_safe('os')
    else:
        os_build = conanfile.settings.get_safe('os_build')
    if os_build is None:  # Assume is the same specified in host settings, not cross-building
        os_build = conanfile.settings.get_safe("os")

    if os_build == "Windows":
        compiler = conanfile.settings.get_safe("compiler")
        sub = conanfile.settings.get_safe("os.subsystem")
        if sub in ("cygwin", "msys2", "msys") or compiler == "qcc":
            return False
        else:
            return True
    return False



import os
import platform

from conans.errors import ConanException
from conans.model.version import Version
from conans.util.runners import check_output_runner


class _OSInfo(object):
    """ Usage:
        (os_info.is_linux) # True/False
        (os_info.is_windows) # True/False
        (os_info.is_macos) # True/False
        (os_info.is_freebsd) # True/False
        (os_info.is_solaris) # True/False

        (os_info.linux_distro)  # debian, ubuntu, fedora, centos...

        (os_info.os_version) # 5.1
        (os_info.os_version_name) # Windows 7, El Capitan

        if os_info.os_version > "10.1":
            pass
        if os_info.os_version == "10.1.0":
            pass
    """

    def __init__(self):
        system = platform.system()
        self.os_version = None

        self.is_linux = system == "Linux"
        self.linux_distro = None
        self.is_msys = system.startswith("MING") or system.startswith("MSYS_NT")
        self.is_cygwin = system.startswith("CYGWIN_NT")
        self.is_windows = system == "Windows" or self.is_msys or self.is_cygwin
        self.is_macos = system == "Darwin"
        self.is_freebsd = system == "FreeBSD"
        self.is_solaris = system == "SunOS"
        self.is_aix = system == "AIX"
        self.is_posix = os.pathsep == ':'

    def _get_linux_distro_info(self):
        import distro
        self.linux_distro = distro.id()
        self.os_version = Version(distro.version())

    @staticmethod
    def get_aix_architecture():
        processor = platform.processor()
        if "powerpc" in processor:
            kernel_bitness = _OSInfo().get_aix_conf("KERNEL_BITMODE")
            if kernel_bitness:
                return "ppc64" if kernel_bitness == "64" else "ppc32"
        elif "rs6000" in processor:
            return "ppc32"

    @staticmethod
    def get_aix_conf(options=None):
        options = " %s" % options if options else ""
        if not _OSInfo().is_aix:
            raise ConanException("Command only for AIX operating system")

        try:
            ret = check_output_runner("getconf%s" % options).strip()
            return ret
        except Exception:
            return None

    @staticmethod
    def get_solaris_architecture():
        # under intel solaris, platform.machine()=='i86pc' so we need to handle
        # it early to suport 64-bit
        processor = platform.processor()
        kernel_bitness, elf = platform.architecture()
        if "sparc" in processor:
            return "sparcv9" if kernel_bitness == "64bit" else "sparc"
        elif "i386" in processor:
            return "x86_64" if kernel_bitness == "64bit" else "x86"

    @staticmethod
    def get_e2k_architecture():
        return {
            "E1C+": "e2k-v4",  # Elbrus 1C+ and Elbrus 1CK
            "E2C+": "e2k-v2",  # Elbrus 2CM
            "E2C+DSP": "e2k-v2",  # Elbrus 2C+
            "E2C3": "e2k-v6",  # Elbrus 2C3
            "E2S": "e2k-v3",  # Elbrus 2S (aka Elbrus 4C)
            "E8C": "e2k-v4",  # Elbrus 8C and Elbrus 8C1
            "E8C2": "e2k-v5",  # Elbrus 8C2 (aka Elbrus 8CB)
            "E12C": "e2k-v6",  # Elbrus 12C
            "E16C": "e2k-v6",  # Elbrus 16C
            "E32C": "e2k-v7",  # Elbrus 32C
        }.get(platform.processor())


def _detected_os():
    if _OSInfo().is_macos:
        return "Macos"
    if _OSInfo().is_windows:
        return "Windows"
    return platform.system()


def _detected_architecture():
    # FIXME: Very weak check but not very common to run conan in other architectures
    machine = platform.machine()
    os_info = _OSInfo()
    arch = None

    if os_info.is_solaris:
        arch = _OSInfo.get_solaris_architecture()
    elif os_info.is_aix:
        arch = _OSInfo.get_aix_architecture()

    if arch:
        return arch

    if "ppc64le" in machine:
        return "ppc64le"
    elif "ppc64" in machine:
        return "ppc64"
    elif "ppc" in machine:
        return "ppc32"
    elif "mips64" in machine:
        return "mips64"
    elif "mips" in machine:
        return "mips"
    elif "sparc64" in machine:
        return "sparcv9"
    elif "sparc" in machine:
        return "sparc"
    elif "aarch64" in machine:
        return "armv8"
    elif "arm64" in machine:
        return "armv8"
    elif "64" in machine:
        return "x86_64"
    elif "86" in machine:
        return "x86"
    elif "armv8" in machine:
        return "armv8"
    elif "armv7" in machine:
        return "armv7"
    elif "arm" in machine:
        return "armv6"
    elif "s390x" in machine:
        return "s390x"
    elif "s390" in machine:
        return "s390"
    elif "sun4v" in machine:
        return "sparc"
    elif "e2k" in machine:
        return _OSInfo.get_e2k_architecture()


def _get_build_os_arch(conanfile):
    """ Returns the value for the 'os' and 'arch' settings for the build context """
    if hasattr(conanfile, 'settings_build'):
        return conanfile.settings_build.get_safe('os'), conanfile.settings_build.get_safe('arch')
    else:
        return conanfile.settings.get_safe('os_build'), conanfile.settings.get_safe('arch_build')


def _get_cross_building_settings(conanfile, self_os=None, self_arch=None):
    os_build, arch_build = _get_build_os_arch(conanfile)
    if not hasattr(conanfile, 'settings_build'):
        # Let it override from outside only if no 'profile_build' is used
        os_build = self_os or os_build or _detected_os()
        arch_build = self_arch or arch_build or _detected_architecture()

    os_host = conanfile.settings.get_safe("os")
    arch_host = conanfile.settings.get_safe("arch")

    return os_build, arch_build, os_host, arch_host



import fnmatch
import os
import textwrap
from collections import OrderedDict

from conans.errors import ConanException
from conans.util.files import save


class _EnvVarPlaceHolder:
    pass


class _Sep(str):
    pass


class _PathSep:
    pass


def environment_wrap_command(filename, cmd, cwd=None):
    assert filename
    filenames = [filename] if not isinstance(filename, list) else filename
    bats, shs = [], []
    for f in filenames:
        full_path = os.path.join(cwd, f) if cwd else f
        if os.path.isfile("{}.bat".format(full_path)):
            bats.append("{}.bat".format(f))
        elif os.path.isfile("{}.sh".format(full_path)):
            shs.append("{}.sh".format(f))
    if bats and shs:
        raise ConanException("Cannot wrap command with different envs, {} - {}".format(bats, shs))

    if bats:
        command = " && ".join('"{}"'.format(b) for b in bats)
        return "{} && {}".format(command, cmd)
    elif shs:
        command = " && ".join(". ./{}".format(f) for f in shs)
        return "{} && {}".format(command, cmd)
    else:
        return cmd


class Environment:
    def __init__(self):
        # TODO: Maybe we need to pass conanfile to get the [conf]
        # It being ordered allows for Windows case-insensitive composition
        self._values = OrderedDict()  # {var_name: [] of values, including separators}

    def __bool__(self):
        return bool(self._values)

    __nonzero__ = __bool__

    def __repr__(self):
        return repr(self._values)

    def vars(self):
        return list(self._values.keys())

    def value(self, name, placeholder="{name}", pathsep=os.pathsep):
        return self._format_value(name, self._values[name], placeholder, pathsep)

    @staticmethod
    def _format_value(name, varvalues, placeholder, pathsep):
        values = []
        for v in varvalues:

            if v is _EnvVarPlaceHolder:
                values.append(placeholder.format(name=name))
            elif v is _PathSep:
                values.append(pathsep)
            else:
                values.append(v)
        return "".join(values)

    @staticmethod
    def _list_value(value, separator):
        if isinstance(value, list):
            result = []
            for v in value[:-1]:
                result.append(v)
                result.append(separator)
            result.extend(value[-1:])
            return result
        else:
            return [value]

    def define(self, name, value, separator=" "):
        value = self._list_value(value, _Sep(separator))
        self._values[name] = value

    def define_path(self, name, value):
        value = self._list_value(value, _PathSep)
        self._values[name] = value

    def unset(self, name):
        """
        clears the variable, equivalent to a unset or set XXX=
        """
        self._values[name] = []

    def append(self, name, value, separator=" "):
        value = self._list_value(value, _Sep(separator))
        self._values[name] = [_EnvVarPlaceHolder] + [_Sep(separator)] + value

    def append_path(self, name, value):
        value = self._list_value(value, _PathSep)
        self._values[name] = [_EnvVarPlaceHolder] + [_PathSep] + value

    def prepend(self, name, value, separator=" "):
        value = self._list_value(value, _Sep(separator))
        self._values[name] = value + [_Sep(separator)] + [_EnvVarPlaceHolder]

    def prepend_path(self, name, value):
        value = self._list_value(value, _PathSep)
        self._values[name] = value + [_PathSep] + [_EnvVarPlaceHolder]

    def save_bat(self, filename, generate_deactivate=False, pathsep=os.pathsep):
        deactivate = textwrap.dedent("""\
            echo Capturing current environment in deactivate_{filename}
            setlocal
            echo @echo off > "deactivate_{filename}"
            echo echo Restoring environment >> "deactivate_{filename}"
            for %%v in ({vars}) do (
                set foundenvvar=
                for /f "delims== tokens=1,2" %%a in ('set') do (
                    if "%%a" == "%%v" (
                        echo set %%a=%%b>> "deactivate_{filename}"
                        set foundenvvar=1
                    )
                )
                if not defined foundenvvar (
                    echo set %%v=>> "deactivate_{filename}"
                )
            )
            endlocal

            """).format(filename=filename, vars=" ".join(self._values.keys()))
        capture = textwrap.dedent("""\
            @echo off
            {deactivate}
            echo Configuring environment variables
            """).format(deactivate=deactivate if generate_deactivate else "")
        result = [capture]
        for varname, varvalues in self._values.items():
            value = self._format_value(varname, varvalues, "%{name}%", pathsep)
            result.append('set {}={}'.format(varname, value))

        content = "\n".join(result)
        save(filename, content)

    def save_ps1(self, filename, generate_deactivate=False, pathsep=os.pathsep):
        # FIXME: This is broken and doesnt work
        deactivate = ""
        capture = textwrap.dedent("""\
            {deactivate}
            """).format(deactivate=deactivate if generate_deactivate else "")
        result = [capture]
        for varname, varvalues in self._values.items():
            value = self._format_value(varname, varvalues, "$env:{name}", pathsep)
            result.append('$env:{}={}'.format(varname, value))

        content = "\n".join(result)
        save(filename, content)

    def save_sh(self, filename, generate_deactivate=False, pathsep=os.pathsep):
        deactivate = textwrap.dedent("""\
            echo Capturing current environment in deactivate_{filename}
            echo echo Restoring variables >> deactivate_{filename}
            for v in {vars}
            do
                value=$(printenv $v)
                if [ -n "$value" ]
                then
                    echo export "$v=$value" >> deactivate_{filename}
                else
                    echo unset $v >> deactivate_{filename}
                fi
            done
            echo Configuring environment variables
            """.format(filename=filename, vars=" ".join(self._values.keys())))
        capture = textwrap.dedent("""\
           {deactivate}
           echo Configuring environment variables
           """).format(deactivate=deactivate if generate_deactivate else "")
        result = [capture]
        for varname, varvalues in self._values.items():
            value = self._format_value(varname, varvalues, "${name}", pathsep)
            if value:
                result.append('export {}="{}"'.format(varname, value))
            else:
                result.append('unset {}'.format(varname))

        content = "\n".join(result)
        save(filename, content)

    def compose(self, other):
        """
        self has precedence, the "other" will add/append if possible and not conflicting, but
        self mandates what to do
        :type other: Environment
        """
        for k, v in other._values.items():
            existing = self._values.get(k)
            if existing is None:
                self._values[k] = v
            else:
                try:
                    index = existing.index(_EnvVarPlaceHolder)
                except ValueError:  # It doesn't have placeholder
                    pass
                else:
                    new_value = existing[:]  # do a copy
                    new_value[index:index + 1] = v  # replace the placeholder
                    # Trim front and back separators
                    val = new_value[0]
                    if isinstance(val, _Sep) or val is _PathSep:
                        new_value = new_value[1:]
                    val = new_value[-1]
                    if isinstance(val, _Sep) or val is _PathSep:
                        new_value = new_value[:-1]
                    self._values[k] = new_value
        return self


class ProfileEnvironment:
    def __init__(self):
        self._environments = OrderedDict()

    def __repr__(self):
        return repr(self._environments)

    def get_env(self, ref):
        """ computes package-specific Environment
        it is only called when conanfile.buildenv is called
        the last one found in the profile file has top priority
        """
        result = Environment()
        for pattern, env in self._environments.items():
            if pattern is None or fnmatch.fnmatch(str(ref), pattern):
                result = env.compose(result)
        return result

    def compose(self, other):
        """
        :type other: ProfileEnvironment
        """
        for pattern, environment in other._environments.items():
            existing = self._environments.get(pattern)
            if existing is not None:
                self._environments[pattern] = environment.compose(existing)
            else:
                self._environments[pattern] = environment

    @staticmethod
    def loads(text):
        result = ProfileEnvironment()
        for line in text.splitlines():
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            for op, method in (("+=", "append"), ("=+", "prepend"),
                               ("=!", "unset"), ("=", "define")):
                tokens = line.split(op, 1)
                if len(tokens) != 2:
                    continue
                pattern_name, value = tokens
                pattern_name = pattern_name.split(":", 1)
                if len(pattern_name) == 2:
                    pattern, name = pattern_name
                else:
                    pattern, name = None, pattern_name[0]

                env = Environment()
                if method == "unset":
                    env.unset(name)
                else:
                    if value.startswith("(path)"):
                        value = value[6:]
                        method = method + "_path"
                    getattr(env, method)(name, value)

                existing = result._environments.get(pattern)
                if existing is None:
                    result._environments[pattern] = env
                else:
                    result._environments[pattern] = env.compose(existing)
                break
            else:
                raise ConanException("Bad env defintion: {}".format(line))
        return result



from conan.tools.gnu.get_cross_building_settings import _get_cross_building_settings


def _cross_building(conanfile, self_os=None, self_arch=None, skip_x64_x86=False):
    ret = _get_cross_building_settings(conanfile, self_os, self_arch)

    build_os, build_arch, host_os, host_arch = ret

    if skip_x64_x86 and host_os is not None and (build_os == host_os) and \
            host_arch is not None and ((build_arch == "x86_64") and (host_arch == "x86") or
                                       (build_arch == "sparcv9") and (host_arch == "sparc") or
                                       (build_arch == "ppc64") and (host_arch == "ppc32")):
        return False

    if host_os is not None and (build_os != host_os):
        return True
    if host_arch is not None and (build_arch != host_arch):
        return True

    return False



import json
import os

from conan.tools import CONAN_TOOLCHAIN_ARGS_FILE
from conan.tools._compilers import use_win_mingw
from conans.util.files import load


class Autotools(object):

    def __init__(self, conanfile):
        """
        FIXME: include_rpath_flags CONAN 2.0 to default True? Could break many packages in center
        """
        self._conanfile = conanfile
        self._win_bash = False
        self._include_rpath_flags = False
        self._os = conanfile.settings.get_safe("os")
        self._os_version = conanfile.settings.get_safe("os.version")
        self._os_sdk = conanfile.settings.get_safe("os.sdk")
        self._os_subsystem = conanfile.settings.get_safe("os.subsystem")
        self._arch = conanfile.settings.get_safe("arch")
        self._build_type = conanfile.settings.get_safe("build_type")
        self._compiler = conanfile.settings.get_safe("compiler")
        self._compiler_version = conanfile.settings.get_safe("compiler.version")
        self._build = None
        self._host = None
        self._target = None
        if os.path.isfile(CONAN_TOOLCHAIN_ARGS_FILE):
            args = json.loads(load(CONAN_TOOLCHAIN_ARGS_FILE))
            self._build = args["build"] if "build" in args else None
            self._host = args["host"] if "host" in args else None
            self._target = args["target"] if "target" in args else None

    def configure(self):
        """
        http://jingfenghanmax.blogspot.com.es/2010/09/configure-with-host-target-and-build.html
        https://gcc.gnu.org/onlinedocs/gccint/Configure-Terms.html
        """
        if not self._conanfile.should_configure:
            return
        configure_dir = "."

        # TODO: Management of PKG_CONFIG_PATH
        # TODO: Implement management of --prefix, bindir, sbindir, libexecdir, libdir, includedir

        cmd = "%s/configure" % configure_dir
        cmd += ' --host=%s' % self._host if self._host else ''
        cmd += ' --build=%s' % self._build if self._build else ''
        cmd += ' --target=%s' % self._target if self._target else ''
        self._conanfile.output.info("Calling:\n > %s" % cmd)
        self._conanfile.run(cmd)

    def make(self):
        """if not self._build_type:
            raise ConanException("build_type setting should be defined.")
        with environment_append(vars or self.vars):
            str_args = args_to_string(args)
            cpu_count_option = (("-j%s" % cpu_count(output=self._conanfile.output))
                                if ("-j" not in str_args and "nmake" not in make_program.lower())
                                else None)
            self._conanfile.run("%s" % join_arguments([make_program, target, str_args,
                                                       cpu_count_option]),
                                win_bash=self._win_bash, subsystem=self.subsystem)"""

        make_program = self._conanfile.conf["tools.gnu:make_program"]
        if make_program is None:
            make_program = "mingw32-make" if use_win_mingw(self._conanfile) else "make"
        # Need to activate the buildenv if existing
        command = make_program
        self._conanfile.run(command)



from conans.client.graph.graph import CONTEXT_HOST, CONTEXT_BUILD
from conans.errors import ConanException
from conans.model.conanfile_interface import ConanFileInterface


class DependencyOrderedSet:
    """ class to give a ["mydep"] access to dependencies, because recipes many times just
    want to access by dependency name, like:
    self.dependencies.requires["openssl"].version
    self.dependencies.build_requires["cmake"].ref
    """
    def __init__(self, deps=None):
        self._deps = deps or []

    def __iter__(self):
        return iter(self._deps)

    def __next__(self):
        return next(self._deps)

    def __getitem__(self, name):
        result = [d for d in self._deps if d.ref.name == name]
        if len(result) > 1:
            raise ConanException("There is more than one dependency")
        if len(result) == 0:
            raise ConanException("No dependency found")
        return result[0]

    def __add__(self, other):
        return DependencyOrderedSet(self._deps + other._deps)


class ConanFileDependencies:

    def __init__(self, node):
        self._node = node

    @property
    def build_requires(self):
        """
        :return: list of immediate direct build_requires
        """
        return DependencyOrderedSet([ConanFileInterface(edge.dst.conanfile)
                                     for edge in self._node.dependencies if edge.build_require])

    @property
    def build_requires_build_context(self):
        """
        :return: list of immediate direct build_requires, on build context.
        FIXME: Why this method? To overcome the legacy use case without 2 profiles where everthing
               is host, otherwise we can receive the same build require twice, one in
               .transitive_host_requires and one in .build_requires
        """
        return DependencyOrderedSet([ConanFileInterface(edge.dst.conanfile)
                                     for edge in self._node.dependencies if edge.build_require and
                                     edge.dst.context == CONTEXT_BUILD])

    @property
    def requires(self):
        """
        :return: list of immediate direct requires, not included build or private ones
        """
        return DependencyOrderedSet([ConanFileInterface(edge.dst.conanfile)
                                     for edge in self._node.dependencies
                                     if not edge.build_require and not edge.private])

    @property
    def host_requires(self):
        """
        :return: list of immediate direct requires and build_requires in the host context
        """
        requires = [ConanFileInterface(edge.dst.conanfile) for edge in self._node.dependencies
                    if not edge.build_require and not edge.private]
        requires.extend([ConanFileInterface(edge.dst.conanfile)
                         for edge in self._node.dependencies
                         if edge.build_require and edge.dst.context == CONTEXT_HOST])
        return DependencyOrderedSet(requires)

    @property
    def transitive_host_requires(self):
        """
        :return: list of direct_host_requires plus all the transitive regular requires of those
        """
        result = []
        next_requires = self.host_requires
        while next_requires:
            new_requires = []
            for require in next_requires:
                if require not in new_requires and require not in result:
                    result.append(require)
                for transitive in require.dependencies.requires:
                    if transitive not in new_requires:
                        new_requires.append(transitive)
            next_requires = new_requires
        return DependencyOrderedSet(result)



import platform

from conan.tools.env import VirtualEnv
from conan.tools.gnu import AutotoolsToolchain, AutotoolsDeps


class AutotoolsGen:
    def __init__(self, conanfile):
        self.toolchain = AutotoolsToolchain(conanfile)
        self.deps = AutotoolsDeps(conanfile)
        self.env = VirtualEnv(conanfile)

    def build_environment(self):
        envtoolchain = self.toolchain.environment()
        envdeps = self.deps.environment()
        build_env = self.env.build_environment()
        build_env.compose(envtoolchain)
        build_env.compose(envdeps)
        return build_env

    def run_environment(self):
        run_env = self.env.run_environment()
        return run_env

    def generate(self):
        build_env = self.build_environment()
        run_env = self.run_environment()
        # FIXME: Use settings, not platform Not always defined :(
        # os_ = self._conanfile.settings_build.get_safe("os")
        if platform.system() == "Windows":
            build_env.save_bat("conanbuildenv.bat")
            run_env.save_bat("conanrunenv.bat")
        else:
            build_env.save_sh("conanbuildenv.sh")
            run_env.save_sh("conanrunenv.sh")

        self.toolchain.generate_args()



import json

from conan.tools import CONAN_TOOLCHAIN_ARGS_FILE
from conan.tools._compilers import architecture_flag, build_type_flags
from conan.tools.env import Environment
from conan.tools.gnu.cross_building import _cross_building
from conan.tools.gnu.get_cross_building_settings import _get_cross_building_settings
from conan.tools.gnu.get_gnu_triplet import _get_gnu_triplet
# FIXME: need to refactor this import and bring to conan.tools
from conans.client.build.cppstd_flags import cppstd_flag_new
from conans.util.files import save


class AutotoolsToolchain:
    def __init__(self, conanfile):
        self._conanfile = conanfile
        build_type = self._conanfile.settings.get_safe("build_type")

        # TODO: compiler.runtime for Visual studio?
        # defines
        self.ndebug = None
        if build_type in ['Release', 'RelWithDebInfo', 'MinSizeRel']:
            self.ndebug = "NDEBUG"
        self.gcc_cxx11_abi = self._cxx11_abi_define()
        self.defines = []

        # cxxflags, cflags
        self.cxxflags = []
        self.cflags = []
        self.ldflags = []
        self.libcxx = self._libcxx()
        self.fpic = self._conanfile.options.get_safe("fPIC")

        # FIXME: This needs to be imported here into conan.tools
        self.cppstd = cppstd_flag_new(self._conanfile.settings)
        self.arch_flag = architecture_flag(self._conanfile.settings)
        # TODO: This is also covering compilers like Visual Studio, necessary to test it (&remove?)
        self.build_type_flags = build_type_flags(self._conanfile.settings)

        self._host = None
        self._build = None
        self._target = None

        if _cross_building(self._conanfile):
            os_build, arch_build, os_host, arch_host = _get_cross_building_settings(self._conanfile)
            self._host = _get_gnu_triplet(os_host, arch_host)
            self._build = _get_gnu_triplet(os_build, arch_build)

    def _rpaths_link(self):
        # TODO: Not implemented yet
        pass

    # TODO: Apple: tools.apple_deployment_target_flag,
    # TODO:  tools.XCRun(self._conanfile.settings).sdk_path
    # TODO: "-arch", tools.to_apple_arch(self._arch)

    def _cxx11_abi_define(self):
        # https://gcc.gnu.org/onlinedocs/libstdc++/manual/using_dual_abi.html
        # The default is libstdc++11, only specify the contrary '_GLIBCXX_USE_CXX11_ABI=0'
        settings = self._conanfile.settings
        libcxx = settings.get_safe("compiler.libcxx")
        if not libcxx:
            return

        compiler = settings.get_safe("compiler.base") or settings.get_safe("compiler")
        if compiler == "gcc":
            if libcxx == 'libstdc++':
                return '_GLIBCXX_USE_CXX11_ABI=0'

    def _libcxx(self):
        settings = self._conanfile.settings
        libcxx = settings.get_safe("compiler.libcxx")
        if not libcxx:
            return

        compiler = settings.get_safe("compiler.base") or settings.get_safe("compiler")

        if compiler in ['clang', 'apple-clang']:
            if libcxx in ['libstdc++', 'libstdc++11']:
                return '-stdlib=libstdc++'
            elif libcxx == 'libc++':
                return '-stdlib=libc++'
        elif compiler == 'sun-cc':
            return ({"libCstd": "-library=Cstd",
                     "libstdcxx": "-library=stdcxx4",
                     "libstlport": "-library=stlport4",
                     "libstdc++": "-library=stdcpp"}.get(libcxx))
        elif compiler == "qcc":
            return "-Y _%s" % str(libcxx)

    def environment(self):
        env = Environment()
        # defines
        if self.ndebug:
            self.defines.append(self.ndebug)
        if self.gcc_cxx11_abi:
            self.defines.append(self.gcc_cxx11_abi)

        if self.libcxx:
            self.cxxflags.append(self.libcxx)

        if self.cppstd:
            self.cxxflags.append(self.cppstd)

        if self.arch_flag:
            self.cxxflags.append(self.arch_flag)
            self.cflags.append(self.arch_flag)
            self.ldflags.append(self.arch_flag)

        if self.build_type_flags:
            self.cxxflags.extend(self.build_type_flags)
            self.cflags.extend(self.build_type_flags)

        if self.fpic:
            self.cxxflags.append("-fPIC")
            self.cflags.append("-fPIC")

        env.append("CPPFLAGS", ["-D{}".format(d) for d in self.defines])
        env.append("CXXFLAGS", self.cxxflags)
        env.append("CFLAGS", self.cflags)
        env.append("LDFLAGS", self.ldflags)
        return env

    def generate(self):
        env = self.environment()
        env.save_sh("conanautotoolstoolchain.sh")
        env.save_bat("conanautotoolstoolchain.bat")
        self.generate_args()

    def generate_args(self):
        args = {"build": self._build,
                "host": self._host,
                "target": self._target}
        args = {k: v for k, v in args.items() if v is not None}
        save(CONAN_TOOLCHAIN_ARGS_FILE, json.dumps(args))
